import datetime
import string
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
import pytest
from numpy.random import randn
from scipy import sparse
from scipy.stats import norm
from sklearn.ensemble import RandomForestClassifier

from mizarlabs.model.bootstrapping import get_ind_matrix
from mizarlabs.model.model_selection import BaseTimeSeriesCrossValidator
from mizarlabs.model.pipeline import MizarFeatureUnion
from mizarlabs.model.pipeline import StrategySignalPipeline
from mizarlabs.static import CLOSE
from mizarlabs.static import EVENT_END_TIME
from mizarlabs.static import FIRST_TIMESTAMP
from mizarlabs.static import LABEL
from mizarlabs.static import SIDE
from mizarlabs.static import TIMESTAMP
from mizarlabs.transformers.microstructural_features.vpin import VPIN
from mizarlabs.transformers.targets.labeling import TripleBarrierMethodLabeling
from mizarlabs.transformers.technical.moving_average import MovingAverageCrossOver
from mizarlabs.transformers.technical.moving_average import (
    MovingAverageCrossOverPredictor,
)
from mizarlabs.transformers.trading.bet_sizing import BET_SIZE
from mizarlabs.transformers.trading.bet_sizing import PREDICTION
from mizarlabs.transformers.trading.bet_sizing import PROBABILITY


def prepare_cv_object(
    cv: BaseTimeSeriesCrossValidator,
    n_samples: int,
    time_shift: str,
    randomlize_times: bool,
):
    X, pred_times, eval_times = create_random_sample_set(
        n_samples=n_samples, time_shift=time_shift, randomize_times=randomlize_times
    )
    cv.X = X
    cv.pred_times = pred_times
    cv.eval_times = eval_times
    cv.indices = np.arange(X.shape[0])


def create_test_df(n_samples: int, n_columns: int, freq: str):
    index = pd.date_range(start="2018-01-01", periods=n_samples, freq=freq)
    data = {
        col: [randn() for _ in range(n_samples)]
        for col in string.ascii_uppercase[:n_columns]
    }
    test_df = pd.DataFrame(data, index=index)
    return test_df


def create_random_sample_set(
    n_samples, time_shift="120m", randomize_times=False, freq="60T"
):
    # Create artificial data
    n_columns = 3
    test_df = create_test_df(n_samples, n_columns, freq)

    # Turn the index into a column labeled 'index'
    test_df = test_df.reset_index()
    if randomize_times:
        n_columns = 1
        # Subtract and adds random time deltas to the index column, to create
        # the prediction and evaluation times
        rand_fact = (
            create_test_df(n_samples, n_columns, freq)
            .reset_index(drop=True)
            .squeeze()
            .iloc[: len(test_df)]
            .abs()
        )
        test_df["index"] = test_df["index"].subtract(
            rand_fact.apply(lambda x: x * pd.Timedelta(time_shift))
        )
        rand_fact = (
            create_test_df(n_samples, n_columns, freq)
            .reset_index(drop=True)
            .squeeze()
            .iloc[: len(test_df)]
            .abs()
        )
        test_df["index2"] = test_df["index"].add(
            rand_fact.apply(lambda x: x * pd.Timedelta(time_shift))
        )
    else:
        test_df["index2"] = test_df["index"].apply(
            lambda x: x + pd.Timedelta(time_shift)
        )
    # Sort the data frame by prediction time
    test_df = test_df.sort_values("index")
    X = test_df[["A", "B", "C"]]
    pred_times = test_df["index"]
    exit_times = test_df["index2"]
    return X, pred_times, exit_times


@pytest.fixture
def time_bar_dataframe() -> pd.DataFrame:
    time_bar_dataframe = pd.read_pickle(
        Path(__file__).parent / "data/bar_dataframe.pkl"
    )
    object_columns = time_bar_dataframe.select_dtypes(include="object").columns
    time_bar_dataframe.loc[:, object_columns] = time_bar_dataframe.loc[
        :, object_columns
    ].astype(float)
    time_bar_dataframe.set_index(
        time_bar_dataframe.index.tz_convert(None), inplace=True
    )

    return time_bar_dataframe


@pytest.fixture
def samples_info_sets(dollar_bar_target_labels: pd.DataFrame):
    return dollar_bar_target_labels[EVENT_END_TIME]


@pytest.fixture
def dollar_bar_labels_and_info(dollar_bar_dataframe):
    # create labels
    tbml = TripleBarrierMethodLabeling(
        num_expiration_bars=5, profit_taking_factor=0.1, stop_loss_factor=0.1
    )
    labels_and_info = tbml.transform(dollar_bar_dataframe)
    return labels_and_info


@pytest.fixture
def dollar_bar_dataframe() -> pd.DataFrame:
    dollar_bar_dataframe = pd.read_pickle(
        Path(__file__).parent / "data/dollar_bar_dataframe.pkl"
    )

    object_columns = dollar_bar_dataframe.select_dtypes(include="object").columns
    dollar_bar_dataframe.loc[:, object_columns] = dollar_bar_dataframe.loc[
        :, object_columns
    ].astype(float)

    dollar_bar_dataframe.set_index(
        pd.to_datetime(
            dollar_bar_dataframe[FIRST_TIMESTAMP],
            utc=True,
            unit="ms",
        ),
        inplace=True,
    )
    dollar_bar_dataframe.index.rename(TIMESTAMP, inplace=True)
    dollar_bar_dataframe.set_index(
        dollar_bar_dataframe.index.tz_convert(None), inplace=True
    )

    return dollar_bar_dataframe


@pytest.fixture(params=[0, 10, 20, 30, 40, 50])
def dataframe_for_bet_sizing_testing_2_classes(request):
    size_factor = 1000

    prob_arr = np.array([0.711, 0.898, 0.992, 0.595, 0.544, 0.775] * size_factor)
    side_arr = np.array([1, 1, -1, 1, -1, 1] * size_factor)
    dates = np.array(
        [
            datetime.datetime(2000, 1, 1) + i * datetime.timedelta(days=1)
            for i in range(len(prob_arr))
        ]
    )

    shift_list = [0.5, 1, 2, 1.5, 0.8, 0.2] * 1000
    shift_dt = np.array([datetime.timedelta(days=d) for d in shift_list])
    dates_shifted = dates + shift_dt
    if request.param != 0:
        dates_shifted[-request.param :] = np.nan
    # Calculate the test statistic and bet size.
    z_test = (prob_arr - 0.5) / (prob_arr * (1 - prob_arr)) ** 0.5
    m_signal = side_arr * (2 * norm.cdf(z_test) - 1)

    df = pd.DataFrame.from_dict(
        {
            PROBABILITY: prob_arr,
            SIDE: side_arr,
            PREDICTION: side_arr,
            EVENT_END_TIME: dates_shifted,
            BET_SIZE: m_signal,
        }
    )
    return df.set_index(dates)


@pytest.fixture(params=[0, 10, 20, 30, 40, 50])
def dataframe_for_bet_sizing_testing_3_classes(request):
    size_factor = 1000

    prob_arr = np.array([0.711, 0.898, 0.992, 0.595, 0.544, 0.775] * size_factor)
    side_arr = np.array([1, 0, -1, 0, -1, 1] * size_factor)
    dates = np.array(
        [
            datetime.datetime(2000, 1, 1) + i * datetime.timedelta(days=1)
            for i in range(len(prob_arr))
        ]
    )

    shift_list = [0.5, 1, 2, 1.5, 0.8, 0.2] * 1000
    shift_dt = np.array([datetime.timedelta(days=d) for d in shift_list])
    dates_shifted = dates + shift_dt
    if request.param != 0:
        dates_shifted[-request.param :] = np.nan
    # Calculate the test statistic and bet size.
    z_test = (prob_arr - 0.5) / (prob_arr * (1 - prob_arr)) ** 0.5
    m_signal = side_arr * (2 * norm.cdf(z_test) - 1)

    df = pd.DataFrame.from_dict(
        {
            PROBABILITY: prob_arr,
            SIDE: side_arr,
            PREDICTION: side_arr,
            EVENT_END_TIME: dates_shifted,
            BET_SIZE: m_signal,
        }
    )
    return df.set_index(dates)


@pytest.fixture(scope="function")
def moving_average_crossover_fast_slow_one_up_crossover():
    index = pd.date_range("2020-01-01", periods=8, freq="d")

    fast = pd.Series([1, 1, 1, 1, 2, 3, 3, 4], index=index, name="fast")
    slow = pd.Series([2, 2, 2, 2, 1, 1, 2, 2], index=index, name="slow")

    return pd.DataFrame({"fast": fast, "slow": slow})


@pytest.fixture(scope="function")
def moving_average_crossover_fast_slow_one_down_crossover():
    index = pd.date_range("2020-01-01", periods=8, freq="d")

    fast = pd.Series([2, 2, 2, 2, 1, 1, 2, 2], index=index, name="fast")
    slow = pd.Series([1, 1, 1, 1, 2, 3, 3, 4], index=index, name="slow")

    return pd.DataFrame({"fast": fast, "slow": slow})


@pytest.fixture(scope="function")
def moving_average_crossover_fast_slow_multiple_crossovers():
    index = pd.date_range("2020-01-01", periods=18, freq="d")

    slow = pd.Series(
        [2, 2, 2, 2, 1, 1, 2, 2, 3, 3, 2, 1, 2, 3, 4, 3, 2, 1], index=index, name="slow"
    )
    fast = pd.Series(
        [1, 1, 1, 1, 2, 3, 3, 4, 2, 1, 2, 2, 3, 2, 1, 2, 3, 4], index=index, name="fast"
    )

    return pd.DataFrame({"fast": fast, "slow": slow})


@pytest.fixture(scope="function")
def moving_average_cross_over_transformer():
    fast = 10
    slow = 40
    return MovingAverageCrossOver(
        fast=fast, slow=slow, column_name=CLOSE, fill_between_crossovers=True
    )


@pytest.fixture(scope="function")
def moving_average_cross_over_predictor():
    fast = 10
    slow = 40
    return MovingAverageCrossOverPredictor(
        fast=fast, slow=slow, column_name=CLOSE, fill_between_crossovers=True
    )


@pytest.fixture
def dollar_bar_target_labels(dollar_bar_dataframe: pd.DataFrame) -> pd.DataFrame:
    tbml = TripleBarrierMethodLabeling(
        num_expiration_bars=10,
        profit_taking_factor=0.2,
        stop_loss_factor=0.2,
    )
    target_labels = tbml.transform(dollar_bar_dataframe[[CLOSE]])
    return target_labels.dropna()


@pytest.fixture
def dollar_bar_ind_matrix(dollar_bar_target_labels: pd.DataFrame) -> np.ndarray:
    indicators_matrix = get_ind_matrix(
        samples_info_sets=dollar_bar_target_labels[EVENT_END_TIME],
        price_bars=dollar_bar_target_labels,
        event_end_time_column_name=EVENT_END_TIME,
    )

    return indicators_matrix


@pytest.fixture
def dollar_bar_ind_matrix_indices(
    dollar_bar_target_labels: pd.DataFrame,
) -> List[pd.Timestamp]:
    _, dollar_bar_ind_matrix_indices = get_ind_matrix(
        samples_info_sets=dollar_bar_target_labels[EVENT_END_TIME],
        price_bars=dollar_bar_target_labels,
        event_end_time_column_name=EVENT_END_TIME,
        return_indices=True,
    )

    return dollar_bar_ind_matrix_indices


@pytest.fixture
def ind_matrix():
    ind_matrix = np.array(
        [
            [1, 0, 0, 0],
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 1, 0, 0],
            [0, 1, 0, 0],
            [0, 1, 1, 0],
            [0, 0, 1, 0],
            [0, 0, 1, 1],
            [0, 0, 0, 1],
        ]
    )
    return sparse.lil_matrix(ind_matrix)


@pytest.fixture
def ind_matrix_csc(ind_matrix):
    return ind_matrix.tocsc()


@pytest.fixture
def X_train_perfect(
    dollar_bar_dataframe: pd.DataFrame, dollar_bar_target_labels: pd.DataFrame
):
    idx = 100
    X_train_perfect = dollar_bar_target_labels[[LABEL]].iloc[:idx]
    return X_train_perfect


@pytest.fixture
def X_train_random(
    dollar_bar_dataframe: pd.DataFrame, dollar_bar_target_labels: pd.DataFrame
):
    idx = 100
    return pd.DataFrame(
        np.random.random(idx),
        index=dollar_bar_target_labels.index[:idx],
        columns=["random_feature"],
    )


@pytest.fixture
def y_train(dollar_bar_dataframe: pd.DataFrame, dollar_bar_target_labels: pd.DataFrame):
    idx = 100
    return dollar_bar_target_labels[LABEL].iloc[:idx]


@pytest.fixture
def X_test_perfect(
    dollar_bar_dataframe: pd.DataFrame, dollar_bar_target_labels: pd.DataFrame
):
    idx = 100
    return dollar_bar_target_labels[[LABEL]].iloc[-idx:]


@pytest.fixture
def X_test_random(
    dollar_bar_dataframe: pd.DataFrame, dollar_bar_target_labels: pd.DataFrame
):
    idx = 100
    return pd.DataFrame(
        np.random.random(idx),
        index=dollar_bar_target_labels.index[-idx:],
        columns=["random_feature"],
    )


@pytest.fixture
def y_test(dollar_bar_dataframe: pd.DataFrame, dollar_bar_target_labels: pd.DataFrame):
    idx = 100
    return dollar_bar_target_labels[[LABEL]].iloc[-idx:]


@pytest.fixture
def time_inhomogeneous_data():
    """
    Creates a sample set consisting in 11 samples at 2h intervals,
    spanning 20h, as well as 10 samples at 59m intervals,
    with the first samples of each group occurring at the same time.
    pred_times and eval_times have the following values:
                pred_times          eval_times
    0  2000-01-01 00:00:00 2000-01-01 01:00:00
    1  2000-01-01 00:00:00 2000-01-01 01:00:00
    2  2000-01-01 00:59:00 2000-01-01 01:59:00
    3  2000-01-01 01:58:00 2000-01-01 02:58:00
    4  2000-01-01 02:00:00 2000-01-01 03:00:00
    5  2000-01-01 02:57:00 2000-01-01 03:57:00
    6  2000-01-01 03:56:00 2000-01-01 04:56:00
    7  2000-01-01 04:00:00 2000-01-01 05:00:00
    8  2000-01-01 04:55:00 2000-01-01 05:55:00
    9  2000-01-01 05:54:00 2000-01-01 06:54:00
    10 2000-01-01 06:00:00 2000-01-01 07:00:00
    11 2000-01-01 06:53:00 2000-01-01 07:53:00
    12 2000-01-01 07:52:00 2000-01-01 08:52:00
    13 2000-01-01 08:00:00 2000-01-01 09:00:00
    14 2000-01-01 08:51:00 2000-01-01 09:51:00
    15 2000-01-01 10:00:00 2000-01-01 11:00:00
    16 2000-01-01 12:00:00 2000-01-01 13:00:00
    17 2000-01-01 14:00:00 2000-01-01 15:00:00
    18 2000-01-01 16:00:00 2000-01-01 17:00:00
    19 2000-01-01 18:00:00 2000-01-01 19:00:00
    20 2000-01-01 20:00:00 2000-01-01 21:00:00
    """

    X1, pred_times1, eval_times1 = create_random_sample_set(
        n_samples=11, time_shift="1H", freq="2H"
    )
    X2, pred_times2, eval_times2 = create_random_sample_set(
        n_samples=10, time_shift="1H", freq="59T"
    )
    data1 = pd.concat([X1, pred_times1, eval_times1], axis=1)
    data2 = pd.concat([X2, pred_times2, eval_times2], axis=1)
    data = pd.concat([data1, data2], axis=0, ignore_index=True)
    data = data.sort_values(by=data.columns[3])
    data = data.reset_index(drop=True)
    X = data.iloc[:, 0:3]
    pred_times = data.iloc[:, 3]
    eval_times = data.iloc[:, 4]

    return X, pred_times, eval_times


@pytest.fixture
def strategy_signal_pipeline_only_primary(bar_feature_generator):
    # create dict structure for primary model feature generator
    feature_transformers_primary_model = {"primary_0": bar_feature_generator}

    pipeline = StrategySignalPipeline(
        feature_transformers_primary_model=feature_transformers_primary_model,
        align_on="primary_0",
        align_how={"primary_0": "mean"},
    )
    return pipeline


@pytest.fixture
def strategy_signal_pipeline_only_primary_fitted(
    strategy_signal_pipeline_only_primary,
    bar_feature_generator,
    x_dict_primary,
    dollar_bar_labels_and_info,
):
    # init dummy models
    primary_model = RandomForestClassifier(random_state=1, n_jobs=-1, max_depth=2)

    X, y = strategy_signal_pipeline_only_primary.create_dataset_primary(
        x_dict_primary, dollar_bar_labels_and_info["label"]
    )

    primary_model.fit(X, y)

    return strategy_signal_pipeline_only_primary.set_primary_model(primary_model)


@pytest.fixture()
def bar_feature_generator_longer_warmup(dollar_bar_dataframe):
    # feature generator
    fast_window = 20
    slow_window = 100
    maco = MovingAverageCrossOver(
        fast=fast_window,
        slow=slow_window,
        column_name=CLOSE,
        fill_between_crossovers=True,
    )
    vpin = VPIN(window=100)
    bar_feature_generator = MizarFeatureUnion(
        [("MACO", maco), ("VPIN", vpin)], n_jobs=-1
    )
    return bar_feature_generator


@pytest.fixture
def bar_feature_generator(dollar_bar_dataframe):
    # feature generator
    fast_window = 10
    slow_window = 20
    maco = MovingAverageCrossOver(
        fast=fast_window,
        slow=slow_window,
        column_name=CLOSE,
        fill_between_crossovers=True,
    )
    vpin = VPIN()
    bar_feature_generator = MizarFeatureUnion(
        [("MACO", maco), ("VPIN", vpin)], n_jobs=-1
    )
    return bar_feature_generator


@pytest.fixture
def strategy_signal_pipeline_with_metalabeling(bar_feature_generator):
    feature_transformers_primary_model = {
        "primary_0": bar_feature_generator,
    }
    feature_transformers_metalabeling_model = {"metalabeling_0": bar_feature_generator}

    pipeline = StrategySignalPipeline(
        feature_transformers_primary_model=feature_transformers_primary_model,
        feature_transformers_metalabeling_model=feature_transformers_metalabeling_model,
        metalabeling_use_predictions_primary_model=False,
        metalabeling_use_proba_primary_model=False,
        align_on="primary_0",
        align_how={"metalabeling_0": "mean"},
    )
    return pipeline


@pytest.fixture
def strategy_signal_pipeline_with_metalabeling_fitted(
    strategy_signal_pipeline_with_metalabeling,
    x_dict_primary_metalabeling,
    dollar_bar_labels_and_info,
):
    primary_model = RandomForestClassifier(random_state=1, n_jobs=-1, max_depth=2)
    metalabeling_model = RandomForestClassifier(random_state=1, n_jobs=-1, max_depth=2)
    (
        X_primary,
        y_primary,
    ) = strategy_signal_pipeline_with_metalabeling.create_dataset_primary(
        x_dict_primary_metalabeling, dollar_bar_labels_and_info["label"]
    )
    primary_model.fit(X_primary, y_primary)
    strategy_signal_pipeline_with_metalabeling.set_primary_model(primary_model)
    (
        X_meta,
        y_meta,
    ) = strategy_signal_pipeline_with_metalabeling.create_dataset_metalabeling(
        x_dict_primary_metalabeling, dollar_bar_labels_and_info["label"]
    )
    metalabeling_model.fit(X_meta, y_meta)
    strategy_signal_pipeline_with_metalabeling.set_metalabeling_model(
        metalabeling_model
    )
    return strategy_signal_pipeline_with_metalabeling


@pytest.fixture
def strategy_signal_pipeline_with_metalabeling_longer_warm_up_primary(
    bar_feature_generator, bar_feature_generator_longer_warmup
):
    feature_transformers_primary_model = {
        "primary_0": bar_feature_generator_longer_warmup,
    }
    feature_transformers_metalabeling_model = {"metalabeling_0": bar_feature_generator}

    pipeline = StrategySignalPipeline(
        feature_transformers_primary_model=feature_transformers_primary_model,
        feature_transformers_metalabeling_model=feature_transformers_metalabeling_model,
        metalabeling_use_predictions_primary_model=False,
        metalabeling_use_proba_primary_model=False,
        align_on="primary_0",
        align_how={"metalabeling_0": "mean"},
    )
    return pipeline


@pytest.fixture
def strategy_signal_pipeline_with_metalabeling_longer_warm_up_primary_fitted(
    strategy_signal_pipeline_with_metalabeling_longer_warm_up_primary,
    x_dict_primary_metalabeling,
    dollar_bar_labels_and_info,
):
    primary_model = RandomForestClassifier(random_state=1, n_jobs=-1, max_depth=2)
    metalabeling_model = RandomForestClassifier(random_state=1, n_jobs=-1, max_depth=2)
    (
        X_primary,
        y_primary,
    ) = strategy_signal_pipeline_with_metalabeling_longer_warm_up_primary.create_dataset_primary(
        x_dict_primary_metalabeling, dollar_bar_labels_and_info["label"]
    )
    primary_model.fit(X_primary, y_primary)
    strategy_signal_pipeline_with_metalabeling_longer_warm_up_primary.set_primary_model(
        primary_model
    )
    (
        X_meta,
        y_meta,
    ) = strategy_signal_pipeline_with_metalabeling_longer_warm_up_primary.create_dataset_metalabeling(
        x_dict_primary_metalabeling, dollar_bar_labels_and_info["label"]
    )
    metalabeling_model.fit(X_meta, y_meta)
    strategy_signal_pipeline_with_metalabeling_longer_warm_up_primary.set_metalabeling_model(
        metalabeling_model
    )
    return strategy_signal_pipeline_with_metalabeling_longer_warm_up_primary


@pytest.fixture
def strategy_signal_pipeline_with_metalabeling_longer_warm_up_metalabeling(
    bar_feature_generator, bar_feature_generator_longer_warmup
):
    feature_transformers_primary_model = {
        "primary_0": bar_feature_generator,
    }
    feature_transformers_metalabeling_model = {
        "metalabeling_0": bar_feature_generator_longer_warmup
    }

    pipeline = StrategySignalPipeline(
        feature_transformers_primary_model=feature_transformers_primary_model,
        feature_transformers_metalabeling_model=feature_transformers_metalabeling_model,
        metalabeling_use_predictions_primary_model=False,
        metalabeling_use_proba_primary_model=False,
        align_on="primary_0",
        align_how={"metalabeling_0": "mean"},
    )
    return pipeline


@pytest.fixture
def strategy_signal_pipeline_with_metalabeling_longer_warm_up_metalabeling_fitted(
    strategy_signal_pipeline_with_metalabeling_longer_warm_up_metalabeling,
    x_dict_primary_metalabeling,
    dollar_bar_labels_and_info,
):
    primary_model = RandomForestClassifier(random_state=1, n_jobs=-1, max_depth=2)
    metalabeling_model = RandomForestClassifier(random_state=1, n_jobs=-1, max_depth=2)
    (
        X_primary,
        y_primary,
    ) = strategy_signal_pipeline_with_metalabeling_longer_warm_up_metalabeling.create_dataset_primary(
        x_dict_primary_metalabeling, dollar_bar_labels_and_info["label"]
    )
    primary_model.fit(X_primary, y_primary)
    strategy_signal_pipeline_with_metalabeling_longer_warm_up_metalabeling.set_primary_model(
        primary_model
    )
    (
        X_meta,
        y_meta,
    ) = strategy_signal_pipeline_with_metalabeling_longer_warm_up_metalabeling.create_dataset_metalabeling(
        x_dict_primary_metalabeling, dollar_bar_labels_and_info["label"]
    )
    metalabeling_model.fit(X_meta, y_meta)
    strategy_signal_pipeline_with_metalabeling_longer_warm_up_metalabeling.set_metalabeling_model(
        metalabeling_model
    )
    return strategy_signal_pipeline_with_metalabeling_longer_warm_up_metalabeling


@pytest.fixture
def x_dict_primary(dollar_bar_dataframe):
    return {"primary_0": dollar_bar_dataframe}


@pytest.fixture
def x_dict_primary_metalabeling(dollar_bar_dataframe):
    return {"primary_0": dollar_bar_dataframe, "metalabeling_0": dollar_bar_dataframe}


@pytest.fixture
def strategy_signal_pipeline_with_metalabeling_and_pred_features(bar_feature_generator):
    # create dict structure for primary model feature generator
    # and metalabeling_model
    feature_transformers_primary_model = {
        "primary_0": bar_feature_generator,
    }
    feature_transformers_metalabeling_model = {"metalabeling_0": bar_feature_generator}

    pipeline = StrategySignalPipeline(
        feature_transformers_primary_model=feature_transformers_primary_model,
        feature_transformers_metalabeling_model=feature_transformers_metalabeling_model,
        metalabeling_use_predictions_primary_model=True,
        metalabeling_use_proba_primary_model=True,
        align_on="primary_0",
        align_how={"metalabeling_0": "mean"},
    )
    return pipeline


@pytest.fixture
def strategy_signal_pipeline_with_metalabeling_and_pred_features_fitted(
    strategy_signal_pipeline_with_metalabeling_and_pred_features,
    x_dict_primary_metalabeling,
    dollar_bar_labels_and_info,
):
    primary_model = RandomForestClassifier(random_state=1, n_jobs=-1, max_depth=2)
    metalabeling_model = RandomForestClassifier(random_state=1, n_jobs=-1, max_depth=2)
    (
        X_primary,
        y_primary,
    ) = strategy_signal_pipeline_with_metalabeling_and_pred_features.create_dataset_primary(
        x_dict_primary_metalabeling, dollar_bar_labels_and_info["label"]
    )
    primary_model.fit(X_primary, y_primary)
    strategy_signal_pipeline_with_metalabeling_and_pred_features.set_primary_model(
        primary_model
    )
    (
        X_meta,
        y_meta,
    ) = strategy_signal_pipeline_with_metalabeling_and_pred_features.create_dataset_metalabeling(
        x_dict_primary_metalabeling, dollar_bar_labels_and_info["label"]
    )
    metalabeling_model.fit(X_meta, y_meta)
    return strategy_signal_pipeline_with_metalabeling_and_pred_features.set_metalabeling_model(
        metalabeling_model
    )


@pytest.fixture
def manual_X_y_primary(manual_X_primary, manual_y_primary):
    manual_y_primary.dropna(inplace=True)
    manual_X_primary = manual_X_primary.loc[manual_y_primary.index]
    return manual_X_primary, manual_y_primary


@pytest.fixture
def manual_X_primary(dollar_bar_dataframe, bar_feature_generator):
    return bar_feature_generator.transform(dollar_bar_dataframe).dropna()


@pytest.fixture
def manual_y_primary(dollar_bar_labels_and_info, manual_X_primary):
    return dollar_bar_labels_and_info.loc[manual_X_primary.index, "label"].dropna()


@pytest.fixture
def manual_primary_model_fitted(manual_X_y_primary):
    return RandomForestClassifier(random_state=1, max_depth=2).fit(
        manual_X_y_primary[0], manual_X_y_primary[1]
    )



import numpy as np
import pandas as pd
from mizarlabs.transformers.utils import check_missing_columns
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from mizarlabs import static


class KyleLambda(BaseEstimator, TransformerMixin):
    """
     Kyle lambda liquidity estimator (p.286).

    See page 286 of Advances in Financial Machine Learning by Marcos Lopez de
    Prado for additional information.

    :param window: The window size
    :type window: int
    :param close_column_name: The name of the close column
    :type close_column_name: str, optional
    :param base_asset_volume_column_name: The name of the base asset volume column
    :type base_asset_volume_column_name: str, optional
    """

    def __init__(
        self,
        window: int = 20,
        close_column_name: str = "close",
        base_asset_volume_column_name: str = "base_asset_volume",
    ):
        self.window = window
        self.close_column_name = close_column_name
        self.base_asset_volume_column_name = base_asset_volume_column_name

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X: pd.DataFrame) -> np.ndarray:
        check_missing_columns(
            X, [self.close_column_name, self.base_asset_volume_column_name]
        )

        close_diff_abs_over_vol = (
            X[self.close_column_name].diff().abs()
            / X[self.base_asset_volume_column_name]
        )
        kyle_lambda = close_diff_abs_over_vol.rolling(self.window).mean()
        return kyle_lambda.values.reshape(-1, 1)


class AmihudLambda(BaseEstimator, TransformerMixin):
    """
    Amihud Lambda liquidity estimator (p.288).

    See page 288 of Advances in Financial Machine Learning by Marcos Lopez de
    Prado for additional information.

    :param window: The window size
    :type window: int
    :param close_column_name: The name of the close column
    :type close_column_name: str, optional
    :param quote_asset_volume_column_name: The name of the quote asset volume column
    :type quote_asset_volume_column_name: str, optional
    """

    def __init__(
        self,
        window: int = 20,
        close_column_name: str = static.CLOSE,
        quote_asset_volume_column_name: str = static.QUOTE_ASSET_VOLUME,
    ):
        self.window = window
        self.close_column_name = close_column_name
        self.quote_asset_volume_column_name = quote_asset_volume_column_name

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X: pd.DataFrame) -> np.ndarray:
        check_missing_columns(
            X, [self.close_column_name, self.quote_asset_volume_column_name]
        )

        abs_log_close_diff = np.abs(np.log(X[self.close_column_name]).diff())
        abs_ret_over_quote_vol = (
            abs_log_close_diff / X[self.quote_asset_volume_column_name]
        )
        return abs_ret_over_quote_vol.rolling(self.window).mean().values.reshape(-1, 1)


class HasbrouckLambda(BaseEstimator, TransformerMixin):
    """
    Hasbrouck Lambda price impact estimator(p.289).

    See page 289 of Advances in Financial Machine Learning by Marcos Lopez de
    Prado for additional information.

    :param window: The window size
    :type window: int
    :param close_column_name: The name of the close column
    :type close_column_name: str, optional
    :param quote_asset_volume_column_name: The name of the quote asset volume column
    :type quote_asset_volume_column_name: str, optional
    """

    def __init__(
        self,
        window: int = 20,
        close_column_name: str = static.CLOSE,
        quote_asset_volume_column_name: str = static.QUOTE_ASSET_VOLUME,
    ):
        self.window = window
        self.close_column_name = close_column_name
        self.quote_asset_volume_column_name = quote_asset_volume_column_name

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X: pd.DataFrame) -> np.ndarray:
        check_missing_columns(
            X,
            [self.close_column_name, self.quote_asset_volume_column_name],
        )

        abs_log_close_diff = np.log(X[self.close_column_name]).diff().abs()
        return (
            (abs_log_close_diff / np.sqrt(X[self.quote_asset_volume_column_name]))
            .rolling(self.window)
            .mean()
        ).values.reshape(-1, 1)



import pandas as pd
from mizarlabs.model.bootstrapping import get_ind_matrix
from mizarlabs.model.bootstrapping import calc_average_uniqueness
from mizarlabs.static import EVENT_END_TIME
from mizarlabs.transformers.utils import check_missing_columns
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin


class AverageUniqueness(BaseEstimator, TransformerMixin):
    """
    Calculates the average uniqueness of samples.
    """

    def __init__(self, event_end_time_column_name: str = EVENT_END_TIME):
        self._event_end_time_column_name = event_end_time_column_name

    def transform(self, X: pd.DataFrame) -> pd.Series:
        check_missing_columns(X, [self._event_end_time_column_name])

        assert ~X[self._event_end_time_column_name].isna().any(), (
            "The expiration barrier should always have a value but "
            f" the indices {', '.join([str(index) for index in X[X[self._event_end_time_column_name].isna()].index])} contains NaNs"
        )
        assert X.index.is_unique, (
            f"Index should be unique but indices"
            f" {', '.join([str(index) for index in X[X.index.duplicated(keep='last')][:3].index])} are duplicated"
        )
        ind_mat_csc = get_ind_matrix(
            X.copy()[self._event_end_time_column_name],
            X.copy(),
            self._event_end_time_column_name,
        ).tocsc()

        average_uniqueness_array = calc_average_uniqueness(ind_mat_csc)
        average_uniqueness = pd.Series(average_uniqueness_array, index=X.index)

        return average_uniqueness



from setuptools import find_packages
from setuptools import setup

__version__ = "0.1.10"

with open("requirements.txt") as f:
    requirements = f.readlines()

with open("README.md", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="mizar-labs",
    version=__version__,
    packages=find_packages(),
    include_package_data=True,
    author="MizarAI",
    author_email="info@mizar.ai",
    install_requires=requirements,
    description="Package for building financial machine learning models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MizarAI/mizar-labs",
    project_urls={
        "Bug Tracker": "https://github.com/MizarAI/mizar-labs/issues",
    },
    python_requires=">=3.8",
)



"""
Implementation of Chu-Stinchcombe-White test
"""
import numpy as np
import pandas as pd
from numba import jit
from numba import njit
from numba import prange

ONE_SIDED_POSITIVE = "one_sided_positive"
ONE_SIDED_NEGATIVE = "one_sided_negative"
TWO_SIDED = "two_sided"


@jit(parallel=True, nopython=True)
def get_test_statistics_array(
    side_test: str, t_indices: np.ndarray, time_series_array: np.ndarray
) -> np.ndarray:
    """Returns an array with the test statistics of
    the Chu-Stinchcombe-White and the critical values

    :param side_test: 'one_sided_positive' or 'one_sided_negative' or 'two_sided'
    :type side_test: str
    :param t_indices: t indices to loop over
    :type t_indices: np.ndarray
    :param time_series_array: Time series to be tested
    :type time_series_array: np.ndarray
    :return: Array with two columns, one for the
             test statistic and one for the critical values
    :rtype: np.ndarray
    """
    b_alpha_5_pct = 4.6  # 4.6 is b_a estimate derived via Monte-Carlo
    test_statistics_array = np.empty((t_indices.shape[0], 2), dtype=np.float64)
    # outer loops goes over all t
    for i in prange(len(t_indices)):

        # compute variance
        t = t_indices[i]
        array_t = time_series_array[:t]
        sigma_squared_t = np.sum(np.square(np.diff(array_t))) / (t - 1)

        # init supremum vals
        max_S_n_t_value = -np.inf
        max_S_n_t_critical_value = np.nan  # Corresponds to c_alpha[n,t]

        y_t = array_t[t - 1]
        # inner loop goes over all n between 1 and t
        for j in prange(len(array_t) - 1):
            # compute test statistic
            n = j + 1
            y_n = time_series_array[j]
            y_t_y_n_diff = get_y_t_y_n_diff(side_test, y_t, y_n)
            S_n_t = y_t_y_n_diff / np.sqrt(sigma_squared_t * (t - n))

            # check if new val is better than supremum
            # if so compute new critical value
            if S_n_t > max_S_n_t_value:
                max_S_n_t_value = S_n_t
                max_S_n_t_critical_value = np.sqrt(b_alpha_5_pct + np.log(t - n))
        # store result of iteration
        test_statistics_array[i, :] = np.array(
            [max_S_n_t_value, max_S_n_t_critical_value], dtype=np.float64
        )
    return test_statistics_array


@njit
def get_y_t_y_n_diff(side_test: str, y_t: np.float64, y_n: np.float64) -> np.float64:
    """Returns the difference between y_t and y_n given
    a test specification.

    :param side_test: 'one_sided_positive' or 'one_sided_negative' or 'two_sided'
    :type side_test: str
    :param y_t: value of the series to be tested at time t
    :type y_t: np.float64
    :param y_n: value of the series to be tested at time n, where n < t
    :type y_n: np.float64
    :return: difference between y_t and y_n given a test specification
    :rtype: np.float64
    """
    if side_test == ONE_SIDED_POSITIVE:
        values_diff = y_t - y_n
    elif side_test == ONE_SIDED_NEGATIVE:
        values_diff = y_n - y_t
    else:
        values_diff = abs(y_t - y_n)
    return values_diff


class ChuStinchcombeWhiteStatTest:
    STATISTIC = "statistic"
    CRITICAL_VALUE = "critical_value"

    def __init__(self, side_test: str):
        """
        Chu-Stinchcombe-White test CUSUM Tests on Levels as
        described on page 250-251 in Advances in Financial Machine Learning
        by Marcos Lopez de Prado.

        :param side_test: 'one_sided_positive' or 'one_sided_negative' or 'two_sided'
        :type side_test: str
        """
        expected_vals = [ONE_SIDED_NEGATIVE, ONE_SIDED_POSITIVE, TWO_SIDED]
        assert (
            side_test in expected_vals
        ), f"Expected one of {expected_vals} got {side_test}..."
        self.side_test = side_test

    def run(self, series_to_test: pd.Series) -> pd.DataFrame:
        """Runs the Chu-Stinchcombe-White stat test and returns
        a dataframe with the test statistics and corresponding critical values.

        :param series_to_test: Time series to be tested.
        :type series_to_test: pd.Series
        :return: DataFrame with results.
        :rtype: pd.DataFrame
        """
        t_indices = np.arange(2, len(series_to_test) + 1)
        S_n_t_array = get_test_statistics_array(
            self.side_test, t_indices, series_to_test.values
        )
        return pd.DataFrame(
            S_n_t_array,
            index=series_to_test.index[1:],
            columns=[self.STATISTIC, self.CRITICAL_VALUE],
        )



import itertools as itt
import math
import numbers
import random
from abc import abstractmethod
from typing import Iterable
from typing import List
from typing import Tuple

import numpy as np
import pandas as pd
from mizarlabs.transformers.targets.labeling import LABEL
from scipy.stats import rv_continuous
from sklearn.model_selection._split import _BaseKFold


class BaseTimeSeriesCrossValidator(_BaseKFold):
    """
    Abstract class for time series cross-validation.

    Time series cross-validation requires each sample has a prediction
    time pred_time, at which the features are used to predict the response,
    and an evaluation time eval_time, at which the response is known and the
    error can be computed.
    Importantly, it means that unlike in standard sklearn cross-validation,
    the samples X, response y, pred_times and eval_times must all be pandas
    dataframe/series having the same index. It is also assumed that the
    samples are time-ordered with respect to the prediction time
    (i.e. pred_times is non-decreasing).

    Parameters:
    ----------
    n_splits : int, default=10
        Number of folds. Must be at least 2.

    """

    def __init__(
        self, n_splits=10, pred_times: pd.Series = None, eval_times: pd.Series = None
    ):

        super().__init__(n_splits, shuffle=False, random_state=None)

        if not isinstance(pred_times, pd.Series):
            raise ValueError("pred_times should be a pandas Series.")
        if not isinstance(eval_times, pd.Series):
            raise ValueError("eval_times should be a pandas Series.")

        self.pred_times = pred_times
        self.eval_times = eval_times
        self.indices = None

    @abstractmethod
    def split(self, X: pd.DataFrame, y: pd.Series = None, groups=None):
        """
        Yield the indices of the train and test sets.

        :param X: pd.DataFrame, shape (n_samples, n_features), required
        :param y: pd.Series
        :param groups:  not used, inherited from _BaseKFold
        :return:
        """
        pass


class CombPurgedKFoldCV(BaseTimeSeriesCrossValidator):
    """
    Purged and embargoed combinatorial cross-validation.

    As described in Advances in financial machine learning,
    Marcos Lopez de Prado, 2018.

    The samples are decomposed into n_groups folds containing equal numbers of
    samples, without shuffling. In each cross validation round, n_test_splits
    folds are used as the test set, while the other folds are used as the train
    set. There are as many rounds as n_test_splits folds among the n_groups
    folds.Each sample should be tagged with a prediction time pred_time and an
    evaluation time eval_time. The split is such that the intervals
    [pred_times, eval_times] associated to samples in the train and test set do
    not overlap. (The overlapping samples are dropped.) In addition,
    an "embargo" period is defined, giving the minimal time between an
    evaluation time in the test set and a prediction time in the training set.
    This is to avoid, in the presence of temporal correlation, a contamination
    of the test set by the train set.

    Parameters:
    ----------
    n_groups : int, default=10
        Number of folds. Must be at least 2.
    n_test_splits : int, default=2
        Number of folds used in the test set. Must be at least 1.
    pred_times : pd.Series, shape (n_samples,), required
        Times at which predictions are made. pred_times.index has to
        coincide with X.index.
    eval_times : pd.Series, shape (n_samples,), required
        Times at which the response becomes available and the error can
        be computed. eval_times.index has to
        coincide with X.index.
    embargo_td : pd.Timedelta, default=0
        Embargo period (see explanations above).
    """

    default_embargo_td = pd.Timedelta(minutes=0)

    def __init__(
        self,
        n_groups=10,
        n_test_splits=2,
        pred_times: pd.Series = None,
        eval_times: pd.Series = None,
        embargo_td: pd.Timedelta = default_embargo_td,
    ):

        self.n_groups = n_groups

        n_splits = int(
            math.factorial(n_groups)
            / (math.factorial(n_test_splits) * math.factorial(n_groups - n_test_splits))
        )

        super().__init__(n_splits, pred_times, eval_times)
        if not isinstance(n_test_splits, numbers.Integral):
            raise ValueError(
                f"The number of test folds must be of "
                f"Integral type. {n_test_splits} of type "
                f"{type(n_test_splits)} was passed."
            )
        n_test_splits = int(n_test_splits)
        if n_test_splits <= 0 or n_test_splits > self.n_groups - 1:
            raise ValueError(
                f"K-fold cross-validation requires at least one "
                f"train/test split by setting n_test_splits"
                f" between 1 and n_groups - 1, "
                f"got n_test_splits = {n_test_splits}."
            )
        self.n_test_splits = n_test_splits
        if not isinstance(embargo_td, pd.Timedelta):
            raise ValueError(
                f"The embargo time should be of type Pandas "
                f"Timedelta. {embargo_td} of type "
                f"{type(embargo_td)} was passed."
            )
        if embargo_td < pd.Timedelta(minutes=0):
            raise ValueError(
                f"The embargo time should be positive, got embargo = {embargo_td}."
            )
        self.embargo_td = embargo_td

    def _test_fold_bounds(self, X: pd.DataFrame, y: pd.Series = None) -> List[Tuple]:
        """
        Calculate the bounds per each fold of the test set.

        :param X: pd.DataFrame, shape (n_samples, n_features), required
        :param y: pd.Series
        :return:
        """
        self.indices = np.arange(X.shape[0])

        # Fold boundaries
        fold_bounds = [
            (fold[0], fold[-1] + 1)
            for fold in np.array_split(self.indices, self.n_groups)
        ]
        # List of all combinations of n_test_splits folds selected to become
        # test sets
        selected_fold_bounds = list(itt.combinations(fold_bounds, self.n_test_splits))
        # In order for the first round to have its whole test set at the end
        # of the dataset
        selected_fold_bounds.reverse()
        return selected_fold_bounds

    def split(
        self, X: pd.DataFrame, y: pd.Series = None, groups=None
    ) -> Iterable[Tuple[np.ndarray, np.ndarray]]:
        """
        Yield the indices of the train and test sets.

        Although the samples are passed in the form of a pandas dataframe,
        the indices returned are position indices, not labels.

        :param X: pd.DataFrame, shape (n_samples, n_features), required
        :param y: pd.Series
        :param groups: not used, inherited from _BaseKFold
        :return:
        """
        selected_fold_bounds = self._test_fold_bounds(X, y)

        for fold_bound_list in selected_fold_bounds:
            # Computes the bounds of the test set, and the corresponding
            # indices
            test_fold_bounds, test_indices = self._compute_test_set(fold_bound_list)
            # Computes the train set indices
            train_indices = self._compute_train_set(test_fold_bounds, test_indices)

            yield train_indices, test_indices

    def _compute_train_set(
        self, test_fold_bounds: List[Tuple[int, int]], test_indices: np.ndarray
    ) -> np.ndarray:
        """
        Compute the position indices of samples in the train set.

        :param test_fold_bounds: List of tuples of position indices
            Each tuple records the bounds of a block of indices in the test
            set.
        :param test_indices: np.ndarray
            A numpy array containing all the indices in the test set.
        :return: np.ndarray
            A numpy array containing all the indices in the train set.
        """
        # As a first approximation, the train set is the complement of the
        # test set
        train_indices = np.setdiff1d(self.indices, test_indices)
        # But we now have to purge and embargo
        for test_fold_start, test_fold_end in test_fold_bounds:
            # Purge
            train_indices = purge(self, train_indices, test_fold_start, test_fold_end)
            # Embargo
            train_indices = embargo(self, train_indices, test_indices, test_fold_end)
        return train_indices

    def _compute_test_set(
        self, fold_bound_list: List[Tuple[int, int]]
    ) -> Tuple[List[Tuple[int, int]], np.ndarray]:
        """
        Compute the indices of the samples in the test set.

        :param fold_bound_list: List of tuples of position indices
            Each tuple records the bounds of the folds belonging to the
            test set.
        :return:
            test_fold_bounds: List of tuples of position indices
                Like fold_bound_list, but with the neighboring folds in the
                test set merged.
            test_indices: np.ndarray
                A numpy array containing the test indices.
        """
        test_indices = np.empty(0)
        test_fold_bounds = []
        for fold_start, fold_end in fold_bound_list:
            # Records the boundaries of the current test split
            if not test_fold_bounds or fold_start != test_fold_bounds[-1][-1]:
                test_fold_bounds.append((fold_start, fold_end))
            # If the current test split is contiguous to the previous one,
            # simply updates the endpoint
            elif fold_start == test_fold_bounds[-1][-1]:
                test_fold_bounds[-1] = (test_fold_bounds[-1][0], fold_end)
            test_indices = np.union1d(
                test_indices, self.indices[fold_start:fold_end]
            ).astype(int)
        return test_fold_bounds, test_indices


def embargo(
    cv: BaseTimeSeriesCrossValidator,
    train_indices: np.ndarray,
    test_indices: np.ndarray,
    test_fold_end: int,
) -> np.ndarray:
    """
    Apply the embargo procedure to part of the train set.

    This amounts to dropping the train set samples whose prediction time occurs
    within self.embargo_dt of the test set sample evaluation times.
    This method applies the embargo only to the part of the training set
    immediately following the end of the test set determined by test_fold_end.


    :param cv: Cross-validation class
        Needs to have the attributes cv.pred_times, cv.eval_times,
        cv.embargo_dt and cv.indices.
    :param train_indices: np.ndarray
        A numpy array containing all the indices of the samples currently
        included in the train set.
    :param test_indices: np.ndarray
        A numpy array containing all the indices of the samples in the test
        set.
    :param test_fold_end: int
        Index corresponding to the end of a test set block.
    :return:
        train_indices: np.ndarray
        The same array, with the indices subject to embargo removed.
    """
    if not hasattr(cv, "embargo_td"):
        raise ValueError(
            "The passed cross-validation object should have a "
            "member cv.embargo_td defining the embargo time."
        )
    last_test_eval_time = cv.eval_times.iloc[test_indices[:test_fold_end]].max()
    min_train_index = len(
        cv.pred_times[cv.pred_times <= last_test_eval_time + cv.embargo_td]
    )
    if min_train_index < cv.indices.shape[0]:
        allowed_indices = np.concatenate(
            (cv.indices[:test_fold_end], cv.indices[min_train_index:])
        )
        train_indices = np.intersect1d(train_indices, allowed_indices)
    return train_indices


def purge(
    cv: BaseTimeSeriesCrossValidator,
    train_indices: np.ndarray,
    test_fold_start: int,
    test_fold_end: int,
) -> np.ndarray:
    """
    Purge part of the train set.

    Given a left boundary index test_fold_start of the test set,
    this method removes from the train set all the
    samples whose evaluation time is posterior to the prediction time of the
    first test sample after the boundary.

    :param cv: Cross-validation class,
        Needs to have the attributes cv.pred_times, cv.eval_times and
        cv.indices.
    :param train_indices: np.ndarray, A numpy array containing all the indices
        of the samples currently included in the train set.
    :param test_fold_start: int, Index corresponding to the start of a test
        set block.
    :param test_fold_end: int, Index corresponding to the end of the same test
        set block.
    :return: train_indices: np.ndarray
        A numpy array containing the train indices purged at test_fold_start.
    """
    time_test_fold_start = cv.pred_times.iloc[test_fold_start]
    # The train indices before the start of the test fold, purged.
    train_indices_1 = np.intersect1d(
        train_indices, cv.indices[cv.eval_times < time_test_fold_start]
    )
    # The train indices after the end of the test fold.
    train_indices_2 = np.intersect1d(train_indices, cv.indices[test_fold_end:])
    return np.concatenate((train_indices_1, train_indices_2))


class LogUniformGen(rv_continuous):
    def _cdf(self, x):
        return np.log(x / self.a) / np.log(self.b / self.a)


def log_uniform(a=1, b=None):
    if b is None:
        b = np.exp(1)
    return LogUniformGen(a=a, b=b, name="logUniform")


def compute_back_test_paths(n_splits: int, n_test_splits: int) -> int:
    """
    Compute the number of backtest paths for the combinatorial crossvalidation.

    As explained in pg. 164 of De Prado book this function calculates
    the number of paths that can be used given the total number of splits
    (`n_splits`) and the test splits (`n_test_splits`).

    :param n_splits: the total number of splits
    :type n_splits: int
    :param n_test_splits: the number of splits used in the test set
    :type n_test_splits: int
    :return: number of the backtest paths
    :rtype: int
    """
    den = 1

    for i in range(1, n_test_splits):
        den = den * (n_splits - i)

    return int(den / math.factorial(n_test_splits - 1))


# TODO: adapt to signal pipeline
def combinatorial_cross_validation_paths(
    X: pd.DataFrame,
    y: pd.DataFrame,
    cv: CombPurgedKFoldCV,
    signal_pipeline,
    signal_pipeline_fit_params,
    label_column_name: str = LABEL,
) -> List[pd.DataFrame]:
    """
    Return the paths for the combinatorial cross validation analysis.

    :param X: Dataframe containing the features
    :type X: pd.DataFrame
    :param y: DataFrame containing the target and target info
    :type y: pd.DataFrame
    :param signal_pipeline: The signal pipeline we want to use for creating features
    :type signal_pipeline:
    :param signal_pipeline_fit_params: Fit params for the signal pipeline
    :type signal_pipeline_fit_params:
    :param cv: Combinatorial purged cv object
    :type cv: CombPurgedKFoldCV

    :return:
    """
    # computing the test bounds per each fold
    test_bounds = cv._test_fold_bounds(X, y[label_column_name])
    bound_pred_dict = {}

    for idx, train_test in enumerate(cv.split(X, y[label_column_name])):
        train_idx = train_test[0]
        test_idx = train_test[1]
        current_test_bounds = test_bounds[idx]

        signal_pipeline.fit(
            X.iloc[train_idx].values,
            y.iloc[train_idx][label_column_name].values,
            **signal_pipeline_fit_params,
        )

        # predict each sub test set
        for test_bound in current_test_bounds:

            # select the indeces of the current sub test set
            part_test = test_idx[
                np.where((test_idx >= test_bound[0]) & (test_idx < test_bound[1]))
            ]

            # predictions on the sub test set
            tmp_df = pd.DataFrame(index=X.index[part_test])
            predictions = signal_pipeline.predict(X.iloc[part_test].values)
            tmp_df["predictions"] = predictions
            proba = signal_pipeline.predict_proba(X.iloc[part_test].values)

            max_proba = np.max(proba, axis=1)
            tmp_df["max_proba"] = max_proba

            tmp_df = pd.merge(tmp_df, y, left_index=True, right_index=True, how="left")

            # adding the sub test set to the dictionary where each key is the
            # index start of the sub test set
            if test_bound[0] not in bound_pred_dict.keys():
                bound_pred_dict[test_bound[0]] = []

            bound_pred_dict[test_bound[0]].append(tmp_df)

    n_paths = compute_back_test_paths(cv.n_splits, cv.n_test_splits)
    paths = []

    # from the sub test computed create paths randomly picking one sub test
    # per start index.
    for _ in range(n_paths):
        path_df = pd.DataFrame()
        for k in bound_pred_dict.keys():
            random.shuffle(bound_pred_dict[k])
            tmp_df = bound_pred_dict[k].pop()
            path_df = path_df.append(tmp_df)

        paths.append(path_df.copy())

    yield paths



"""
Explosiveness tests: Chow-Type Dickey-Fuller Test
"""
from typing import Tuple

import numpy as np
import pandas as pd
from numba import jit
from numba import njit
from numba import prange


@jit(parallel=True, nopython=True)
def get_dfc_array(
    indices: np.ndarray, y_diff: np.ndarray, y_lag: np.ndarray
) -> np.ndarray:
    """Returns the Chow-Type Dickey-Fuller t-values.

    :param indices: Indices to iterate over
    :type indices: np.ndarray
    :param y_diff: Differenced time series
    :type y_diff: np.ndarray
    :param y_lag: Lagged time series
    :type y_lag: np.ndarray
    :return: Array with t-values.
    :rtype: np.ndarray
    """
    dfc_array = np.empty_like(indices, dtype=np.float64)
    for i in prange(len(indices)):
        dummy_var = np.ones_like(y_lag)
        dummy_var[: indices[i]] = 0  # D_t* indicator: before t* D_t* = 0
        X = y_lag * dummy_var
        beta_hat, beta_var = get_beta_and_beta_var(
            X,
            y_diff,
        )
        dfc_array[i] = beta_hat[0, 0] / np.sqrt(beta_var[0, 0])
    return dfc_array


@njit
def get_beta_and_beta_var(X: np.ndarray, y: np.ndarray) -> Tuple[float]:
    """
    Returns the OLS estimates of the coefficients and the variance of the coefficients.

    :param X: Matrix with features values
    :type X: pd.DataFrame
    :param y: Series with target values.
    :type y: pd.Series
    :return: Tuple with the coefficients and the variance
                of the coefficients estimate.
    :rtype: Tuple[np.ndarray, np.ndarray]
    """
    Xy = np.dot(X.T, y)
    XX = np.dot(X.T, X)
    XX_inv = np.linalg.inv(XX)
    beta_hat = np.dot(np.ascontiguousarray(XX_inv), Xy)
    err = y - np.dot(X, beta_hat)
    beta_hat_var = np.dot(err.T, err) / (X.shape[0] - X.shape[1]) * XX_inv
    return beta_hat, beta_hat_var


class SupremumDickeyFullerChowStatTest:
    def __init__(self, min_num_samples: int = 20):
        """
        Chow-Type Dickey-Fuller Test statistics as
        described on page 251-252 in Advances in Financial
        Machine Learning by Marcos Lopez de Prado.

        :param min_num_samples: min. no. of samples for the dummy variable
                                in the test specification to ensure enough
                                ones and zeros, defaults to 20
        :type min_num_samples: int, optional
        """
        self.min_num_samples = min_num_samples

    def run(self, series_to_test: pd.Series) -> pd.Series:
        indices = np.arange(
            self.min_num_samples, series_to_test.shape[0] - self.min_num_samples
        )
        series_diff = series_to_test.diff().dropna()
        series_lag = series_to_test.shift(1).dropna()
        dfc_array = get_dfc_array(
            indices,
            np.ascontiguousarray(series_diff.values.reshape(-1, 1), dtype=np.float64),
            np.ascontiguousarray(series_lag.values.reshape(-1, 1), dtype=np.float64),
        )
        return pd.Series(
            dfc_array,
            index=series_to_test.index.values[
                self.min_num_samples : series_to_test.shape[0] - self.min_num_samples
            ],
        )
